import time
import numpy

img_size = 500

def mandelbrot( h,w, maxit=20 ):
    """Returns an image of the Mandelbrot fractal of size (h,w)."""
    y,x = numpy.ogrid[ -1.4:1.4:h*1j, -2:0.8:w*1j ]
    c = x+y*1j
    z = c
    divtime = maxit + numpy.zeros(z.shape, dtype=int)
    for i in range(maxit):
        z = z**2 + c
        diverge = z*numpy.conj(z) > 2**2            # who is diverging
        div_now = diverge & (divtime==maxit)  # who is diverging now
        divtime[div_now] = i                  # note when
        z[diverge] = 2
    return divtime                  

img = mandelbrot(img_size,img_size,maxit = 10 + int( 100 * numpy.random.rand()))


def get_data_names():
    return ['image','intensity']

def get_metadata_names():
    return ['image_roi']
    
def setup():
    msg = "Setup done"
    return msg
	
def stop():
    msg = "Stop done"
    return msg
	
def action(val=None):
    print "Action",val,str(time.asctime())
    result = {}
    img = mandelbrot(img_size,img_size,maxit = 10 + int( 100 * numpy.random.rand()))
    result['image'] = img
    result['intensity'] = img.max()
    result['playerlog'] = "Image created"
    return result

def metadata_action(val=None):
    print "Metadata Action",val,str(time.asctime())
    result = {}
    result['image_roi'] = [0,0,img_size-1,img_size-1]
    return result
