import time
import numpy

img_size = 500

def mandelbrot( h,w, maxit=20 ):
    """Returns an image of the Mandelbrot fractal of size (h,w)."""
    y,x = numpy.ogrid[ -1.4:1.4:h*1j, -2:0.8:w*1j ]
    c = x+y*1j
    z = c
    divtime = maxit + numpy.zeros(z.shape, dtype=numpy.int16)
    for i in range(maxit):
        z = z**2 + c
        diverge = z*numpy.conj(z) > 2**2            # who is diverging
        div_now = diverge & (divtime==maxit)  # who is diverging now
        divtime[div_now] = i                  # note when
        z[diverge] = 2
    return divtime                  

img = mandelbrot(img_size,img_size,maxit = 10 + int( 100 * numpy.random.rand()))


def get_data_names():
    return ['mandelbrot','random']

def get_metadata_names():
    return ['mandelbrot_roi']
    
def setup():
    msg = "Setup done"
    return msg

def stop():
    msg = "Stop done"
    return msg

def action(val=None):
    print("Action",val,str(time.asctime()))
    result = {}
    result['mandelbrot'] = img
    result['random'] = int( 100 * numpy.random.rand())
    result['playerlog'] = "Image created"
    return result

def freerun(cmds_queue, data_queue):
    print("Freerun",str(time.asctime()))
    messages = []
    go_on = True
    daq_trg = -1
    my_trg = -1
    max_trg = 0
    result = {}
    while go_on and (my_trg < max_trg):
        while not cmds_queue.empty():
            try:
                item = cmds_queue.get_nowait()
                messages.append(item)
                cmds_queue.task_done()
            except:
                pass
        while messages:
            msg = messages.pop(0)
            print("This is the message received: ", msg)
            if 'abort' in msg:
                return
            elif 'done' in msg:
                max_trg = daq_trg
            elif 'trg:' in msg:
                daq_trg = int(msg.split('trg:')[-1])
                max_trg = daq_trg + 1
        # All the messages from the DonkiPlayer are read
        if my_trg < daq_trg:
            result['mandelbrot'] = mandelbrot(img_size,img_size,maxit = 10 + int( 100 * numpy.random.rand()))
            result['random'] = int( 100 * numpy.random.rand())
            my_trg += 1
            data_msg = (my_trg, result)
            data_queue.put(data_msg)
        else:
            time.sleep(0.01)

def metadata_action(val=None):
    print("Metadata Action",val,str(time.asctime()))
    result = {}
    result['mandelbrot_roi'] = [0,0,img_size-1,img_size-1]
    return result
