import time
import threading
import os,sys
from circularBuffer import circularBuffer
import socket
import DonkiOrchestraLib
import traceback
import numpy
try:
    from Queue import Queue
except:
    from queue import Queue

THREAD_DELAY_SEC = 0.1
DATABUFFERSIZE = 10
DEBUG = True

class processThread(threading.Thread):

#-----------------------------------------------------------------------------------
#    __init__
#-----------------------------------------------------------------------------------
    def __init__(self, pub_tag, info_server = 'localhost:50010', action_code = 'scripts/tangotest_daq.py'):
        threading.Thread.__init__(self)
        self._alive = True
        self._seq_started = False
        self._state = 'STANDBY'
        self.trg_idx = 0
        self.Director = None
        self.trigger_event_id = 0
        self.databuffers = {}
        self.datalock = threading.Lock()
        self.director_url = ''
        #
        self.info_server = info_server
        self.pub_tag = pub_tag.lower()
        self.ActionCode = action_code
        self.prio = -1
        self.lastlog = ''
        self.code_module = None
        self.onlyAck = True
        self.data_names = []
        self.metadata_names = []
        #
        self.freerun = False
        self.freerun_cmds_queue = Queue(maxsize = 100)
        self.freerun_data_queue = Queue(maxsize = 100)
        self.freerun_thread = None
        #
        self.load_action_module()
        #
        self.zcc = DonkiOrchestraLib.CommunicationClass(self.pub_tag)

#-----------------------------------------------------------------------------------
#    load_action_module :
#
#-----------------------------------------------------------------------------------
    def load_action_module(self):
        if self.code_module is not None:
            del self.code_module
        #
        if len(self.ActionCode) and self.ActionCode.strip().endswith('.py'):
            scriptfile = self.ActionCode.strip()
            if os.path.isfile(scriptfile):
                f = open(scriptfile)
                content = f.readlines()
                f.close()
                self.code_module = self.importCode('\n'.join(content),"action_code")
            else:
                self._state = 'ALARM'
        else:
            self.code_module = self.importCode('\n'.join(self.dev.ActionCode),"action_code")
        #
        code_ok = True
        for method_name in ['setup','stop']:
            if not hasattr(self.code_module,method_name) or not callable(getattr(self.code_module,method_name)):
                code_ok = False
                self._state = 'ALARM'
        if not code_ok:
            return
        try:
            self.data_names = self.code_module.get_data_names()
        except:
            self.data_names = []
        try:
            self.metadata_names = self.code_module.get_metadata_names()
        except:
            self.metadata_names = []
        #
        self.onlyAck = (self.data_names == [])
        if not self.onlyAck:
            for dname in self.data_names:
                self.databuffers[dname.lower()] = circularBuffer(DATABUFFERSIZE)

#-----------------------------------------------------------------------------------
#    exec_action :
#
#-----------------------------------------------------------------------------------
    def exec_action(self, trg):
        if DEBUG:
            t0 = time.time()
        # Send metadata info if necessary
        try:
            if len(self.metadata_names):
                retval = self.do_meta_stuff()
                if retval is not None:
                    for resname in retval:
                        if resname in self.metadata_names:
                            self.zcc.publish_metadata(resname.lower(),trg,retval[resname])
        except:
            if DEBUG:
                traceback.print_exc()
            self.zcc.publish_log(self.pub_tag, trg, trg, str(traceback.format_exc()))
        # Perfrom Player Action
        try:
            retval = self.do_stuff()
            #
            if not self.onlyAck and retval != None:
                for resname in retval:
                    if resname in self.data_names:
                        self.datalock.acquire()
                        self.databuffers[resname.lower()].push(trg, retval[resname])
                        self.datalock.release()
                    elif resname.lower() == 'playerlog':
                        self.lastlog = time.asctime() + " " + retval[resname]
                        self.zcc.publish_log(self.pub_tag, trg, trg, retval[resname])
                        if DEBUG:
                            print(self.lastlog)
                        continue
                    if isinstance(retval[resname], list):
                        self.push_data(resname.lower(),trg,trg,retval[resname])
                    elif isinstance(retval[resname], numpy.ndarray):
                        if (retval[resname]).ndim == 1:
                            resdata = (retval[resname]).reshape(1,(retval[resname]).shape[0])
                        elif (retval[resname]).ndim == 2:
                            resdata = (retval[resname]).reshape(1,(retval[resname]).shape[0],(retval[resname]).shape[1])
                        self.push_data(resname.lower(),trg,trg,resdata)
                    else:
                        self.push_data(resname.lower(),trg,trg,[retval[resname]])
            #
            self.zcc.publish_ack('ack',trg,trg)
        except:
            if DEBUG:
                traceback.print_exc()
            self.zcc.publish_log(self.pub_tag, trg, trg, "ERROR: " + str(traceback.format_exc()))
            self.zcc.publish_ack('ack',trg,trg)
        if DEBUG:
            print("Delay: %g ms " % ((time.time()-t0) * 1000))

#-----------------------------------------------------------------------------------
#    check_freerun_data:
#
#-----------------------------------------------------------------------------------
    def check_freerun_data(self):
        try:
            results = []
            while not self.freerun_data_queue.empty():
                results.append(self.freerun_data_queue.get_nowait())
                self.freerun_data_queue.task_done()
            if self.onlyAck or len(results) == 0:
                return
            for res in results:
                trg = res[0]
                retval = res[1]
                for resname in retval.keys():
                    if resname in self.data_names:
                        self.datalock.acquire()
                        self.databuffers[resname.lower()].push(trg, retval[resname])
                        self.datalock.release()
                    elif resname.lower() == 'playerlog':
                        self.lastlog = time.asctime() + " " + retval[resname]
                        self.zcc.publish_log(self.pub_tag, trg, trg, retval[resname])
                        if DEBUG:
                            print(self.lastlog)
                        continue
                    if isinstance(retval[resname], list):
                        self.push_data(resname.lower(),trg,trg,retval[resname])
                    elif isinstance(retval[resname], numpy.ndarray):
                        if (retval[resname]).ndim == 1:
                            resdata = (retval[resname]).reshape(1,(retval[resname]).shape[0])
                        elif (retval[resname]).ndim == 2:
                            resdata = (retval[resname]).reshape(1,(retval[resname]).shape[0],(retval[resname]).shape[1])
                        self.push_data(resname.lower(),trg,trg,resdata)
                    else:
                        self.push_data(resname.lower(),trg,trg,[retval[resname]])
        except:
            if DEBUG:
                traceback.print_exc()
            self.zcc.publish_log(self.pub_tag, trg, trg, "ERROR: " + str(traceback.format_exc()) )

#-----------------------------------------------------------------------------------
#    update_connection_info
#-----------------------------------------------------------------------------------
    def update_connections(self):
        info_host = self.info_server.split(":")[0]
        info_port = int(self.info_server.split(":")[1])

        try:
            info_sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            info_sock.connect((info_host, info_port))
            # Retrieve director URL
            msg = "get donkidirector\r\n"
            info_sock.send(msg.encode())
            reply = info_sock.recv(2048).strip()
            if reply == '':
                raise RuntimeError("socket connection broken")
            dd_url = eval(reply)[0]['data']
            if dd_url != self.director_url:
                self.zcc.sub_socks.clear()
                self.zcc.create_sub_socket('donkidirector',dd_url)
                # Send my PUB URL
                msg = "set donkiplayers "+self.zcc.pub_tag+" "+self.zcc.my_pub_socket_info()+"\r\n"
                info_sock.send(msg.encode())
                reply = info_sock.recv(2048).strip()
                info_sock.close()
                if reply == '':
                    raise RuntimeError("socket connection broken")
                self.director_url = dd_url
        except:
            if 'Connection refused' in str(traceback.format_exc()):
                print("Information server not responding (%s:%d)" % (info_host, info_port))
            else:
                traceback.print_exc()
            return False
        return True



#-----------------------------------------------------------------------------------
#    push_data :
#
#-----------------------------------------------------------------------------------
    def push_data(self, data_name, trg_start, trg_stop, data_value):
        if DEBUG:
            print(data_name, trg_start, trg_stop, data_value)
        self.zcc.publish_data(data_name,trg_start,trg_stop,data_value)


#-----------------------------------------------------------------------------------
#    get_data :	- data_name: name of the data attributef
#		- params: [0,N] read last N samples, [1,bn_start,bn_stop] read from bn_start to bn_stop
#-----------------------------------------------------------------------------------
    def get_data(self, data_name_in, params):
        data_name = data_name_in.lower()
        #
        if data_name not in self.databuffers:
            raise RuntimeError("Wrong data_name")
            #
            self.datalock.acquire()
            #
        if (params[0] == 0) and (len(params) > 1):
            data = self.databuffers[data_name].get_last_data(params[1])
        elif (params[0] == 1) and (len(params) > 2):
            data = self.databuffers[data_name].get_data_chunk(params[1],params[2])
        else:
            self.datalock.release()
            raise RuntimeError("Wrong command parameters")
        #
        self.datalock.release()
        if len(data) == 0:
            raise RuntimeError("Requested data unavailable")
        return data

#-----------------------------------------------------------------------------------
#    importCode
#-----------------------------------------------------------------------------------
    def importCode(self, code, name, add_to_sys_modules=0):
        """
        Import dynamically generated code as a module. code is the
        object containing the code (a string, a file handle or an
        actual compiled code object, same types as accepted by an
        exec statement). The name is the name to give to the module,
        and the final argument says wheter to add it to sys.modules
        or not. If it is added, a subsequent import statement using
        name will return this module. If it is not added to sys.modules
        import will try to load it in the normal fashion.

        import foo

        is equivalent to

        foofile = open("/path/to/foo.py")
        foo = importCode(foofile,"foo",1)

        Returns a newly generated module.
        """
        import sys,imp

        module = imp.new_module(name)
        if sys.version_info[0] == 2:
            exec(code) in module.__dict__
        else:
            exec(code, module.__dict__)

        if add_to_sys_modules:
        	sys.modules[name] = module

        return module


#-----------------------------------------------------------------------------------
#    do_meta_stuff
#-----------------------------------------------------------------------------------
    def do_meta_stuff(self):
        trg = self.last_trg
        if self._state != 'ON' or (self.metadata_period < 0) or (trg % self.metadata_period) != 1:
            return None
        retval = self.code_module.metadata_action(trg)
        return retval


#-----------------------------------------------------------------------------------
#    do_stuff
#-----------------------------------------------------------------------------------
    def do_stuff(self):
        if self._state != 'ON':
            return None
        #
        trg = self.last_trg
        if self.freerun:
            self.freerun_cmds_queue.put("trg:%d" % trg)
            return None
        else:
            retval = self.code_module.action(trg)
            return retval


#-----------------------------------------------------------------------------------
#    do_start
#-----------------------------------------------------------------------------------
    def do_start(self):
        log_msg = None
        if self._state != 'ON':
            return (False,log_msg)
        # Reload code
        try:
            self.load_action_module()
        except:
            log_msg = "ERROR: " + str(traceback.format_exc())
            self.lastlog = log_msg
            traceback.print_exc()
            return (False,log_msg)
        if not hasattr(self.code_module,'setup') or not callable(getattr(self.code_module,'setup')):
            return (True,log_msg)
        try:
            ret = self.code_module.setup()
            if ret != None:
                log_msg = str(ret)
                self.lastlog = log_msg
                if DEBUG:
                    print( self.lastlog )
            if self.freerun:
                while not self.freerun_data_queue.empty():
                    self.freerun_data_queue.get_nowait()
                if self.freerun_thread is not None and self.freerun_thread.is_alive():
                    self.freerun_cmds_queue.put('abort')
                    t0 = time.time()
                    while self.freerun_thread.is_alive():
                        time.sleep(0.1)
                        if (time.time() - t0) > 3:
                            log_msg = "ERROR: FREERUN BUSY"
                            return (False,log_msg)
                fargs = (self.freerun_cmds_queue,self.freerun_data_queue,)
                self.freerun_thread = threading.Thread(target=self.code_module.freerun, args=fargs)		
                self.freerun_thread.start()
            return (True,log_msg)
        except:
            log_msg = "ERROR: " + str(traceback.format_exc())
            self.lastlog = log_msg
            traceback.print_exc()
        return (False,log_msg)


#-----------------------------------------------------------------------------------
#    do_stop
#-----------------------------------------------------------------------------------
    def do_stop(self):
        log_msg = None
        if not hasattr(self.code_module,'stop') or not callable(getattr(self.code_module,'stop')):
            return (True,log_msg)
        try:
            if self.freerun:
                if self.freerun_thread is not None and self.freerun_thread.is_alive():
                    self.freerun_cmds_queue.put('done')
            else:
                ret = self.code_module.stop()
                if ret != None:
                    log_msg = str(ret)
                    self.lastlog = log_msg
                    if DEBUG:
                        print( self.lastlog )
            return (True,log_msg)
        except:
            log_msg = "ERROR: " + str(traceback.format_exc())
            self.lastlog = log_msg
            traceback.print_exc()
        return (False,log_msg)

#-----------------------------------------------------------------------------------
#    do_pause
#-----------------------------------------------------------------------------------
    def do_pause(self, vaule_in):
        log_msg = None
        if not hasattr(self.code_module,'pause') or not callable(getattr(self.code_module,'pause')):
            return (True,log_msg)
        try:
            ret = self.code_module.pause(vaule_in)
            if ret != None:
                log_msg = str(ret)
                self.lastlog = log_msg
                if DEBUG:
                    print(self.lastlog)
            return (True,log_msg)
        except:
            log_msg = "ERROR: " + str(traceback.format_exc())
            self.lastlog = log_msg
            traceback.print_exc()
        return (False,log_msg)


#-----------------------------------------------------------------------------------
#    run
#-----------------------------------------------------------------------------------
    def run(self):
        #
        self.update_connections()
        #
        last_msg_time = time.time()
        self.last_trg = 0
        while self._alive:
            if self.freerun:
                # Check freerun data queue if needed
                self.check_freerun_data()
            if not self._seq_started:
                # Check for new messages from Director
                dd_msg = self.zcc.wait_message(['donkidirector'])
                try:
                    if 'donkidirector' in dd_msg:
                        last_msg_time = time.time()
                        new_msg = dd_msg['donkidirector']
                        topic = new_msg[0].lower()
                        if topic == 'info' and new_msg[1] == self.pub_tag:
                            time.sleep(0.05)
                            # Reload code
                            try:
                                self.load_action_module()
                            except:
                                log_msg = "ERROR: " + str(traceback.format_exc())
                                self.lastlog = log_msg
                                traceback.print_exc()
                            self.zcc.publish_info(self.prio,self.freerun,self.data_names,self.metadata_names)
                        elif topic == 'playerlog' and new_msg[1] == self.pub_tag:
                            time.sleep(0.1)
                            self.zcc.publish_log(self.pub_tag, -1, -1, self.lastlog)
                        elif topic == 'start' and new_msg[1] == self.pub_tag:
                            if len(new_msg) > 2:
                                self.metadata_period = int(new_msg[2])
                            self._seq_started = True
                        elif topic == 'priority' and new_msg[1] == self.pub_tag:
                            self.prio = new_msg[2]
                            self.zcc.publish_ack('priority',1,1)
                        elif topic == 'freerun' and new_msg[1] == self.pub_tag:
                            self.freerun = new_msg[2]
                            self.zcc.publish_ack('freerun',1,1)
                        elif topic == 'stop' and new_msg[1] == self.pub_tag:
                            (ok, log_msg) = self.do_stop()
                            if ok:
                                self.zcc.publish_ack(topic,1,1)
                            else:
                                self.zcc.publish_ack(topic,-1,-1)
                            if log_msg is not None:
                                self.zcc.publish_log(self.pub_tag, -1, -1, log_msg)
                        elif topic == 'pause' and new_msg[1] == self.pub_tag:
                            value_in = True
                            if len(new_msg) > 2:
                                value_in = bool(new_msg[2])
                            (ok, log_msg) = self.do_pause(value_in)
                            if ok:
                                self.zcc.publish_ack(topic,1,1)
                            else:
                                self.zcc.publish_ack(topic,-1,-1)
                            if log_msg is not None:
                                self.zcc.publish_log(self.pub_tag, -1, -1, log_msg)
                        elif topic == 'trigger':
                            trg = new_msg[1]
                            prio = new_msg[2]
                            if (self.prio >= 0) and ((trg != -1) or (prio != -1)):
                                print("IT'S NOT A DIRECTOR PING TRIGGER!")
                            self.zcc.publish_ack('ack',-1,-1)
                    #
                    if time.time() - last_msg_time > 5:
                        # In the last 5 seconds we did not receive any message...
                        # check director URL
                        self.update_connections()
                        last_msg_time = time.time()
                except:
                    traceback.print_exc()
            else:
                self.trg_idx = 0
                self._state = 'ON'
                self.last_trg = 0
                (ok, log_msg) = self.do_start()
                if ok:
                    # Acknowledge Director that we started
                    self.zcc.publish_ack('start',1,1)
                    if log_msg is not None:
                        self.zcc.publish_log(self.pub_tag, self.last_trg, self.last_trg, log_msg)
                else:
                    # Acknowledge Director that there was an error
                    self.zcc.publish_ack('start',-1,-1)
                    if log_msg is not None:
                        self.zcc.publish_log(self.pub_tag, self.last_trg, self.last_trg, log_msg)
                    self._state = 'STANDBY'
                    self._seq_started = False
                    continue
                while (self._seq_started) and (self._state == 'ON') and (self._alive):
                    try:
                        if self.freerun:
                            # Check freerun data queue if needed
                            self.check_freerun_data()
                        # Check for new messages from Director
                        dd_msg = self.zcc.wait_message(['donkidirector'])
                        if 'donkidirector' in dd_msg:
                            last_msg_time = time.time()
                            new_msg = dd_msg['donkidirector']
                            topic = new_msg[0].lower()
                            if topic == 'trigger':
                                trg = new_msg[1]
                                prio = new_msg[2]
                                if prio == trg == -1:
                                    # It's a DonkiDirectror (stanby) check trigger
                                    # It means that the sequence has been stopped
                                    self._seq_started = False
                                    continue
                                if prio == self.prio and self.last_trg == (trg - 1):
                                    self.last_trg = trg
                                    self.exec_action(trg)
                            elif topic == 'stop':
                                name = str(new_msg[1]).lower()
                                if name == self.pub_tag:
                                    (ok,log_msg) = self.do_stop()
                                    # Acknowledge Director that we've stopped
                                    if ok:
                                        self.zcc.publish_ack(topic,1,1)
                                    else:
                                        self.zcc.publish_ack(topic,-1,-1)
                                    if log_msg is not None:
                                        self.zcc.publish_log(self.pub_tag, self.last_trg, self.last_trg, log_msg)
                                    self._seq_started = False
                            elif topic == 'pause':
                                name = str(new_msg[1]).lower()
                                if name == self.pub_tag:
                                    value_in = True
                                    if len(new_msg) > 2:
                                        value_in = bool(new_msg[2])
                                    (ok, log_msg) = self.do_pause(value_in)
                                    if ok:
                                        # Acknowledge Director that we did it
                                        self.zcc.publish_ack(topic,1,1)
                                    else:
                                        self.zcc.publish_ack(topic,-1,-1)
                                    if log_msg is not None:
                                        self.zcc.publish_log(self.pub_tag, self.last_trg, self.last_trg, log_msg)
                            elif topic == 'start':
                                name = str(new_msg[1]).lower()
                                if name == self.pub_tag:
                                    (ok, log_msg) = self.do_start()
                                    if ok:
                                        # Acknowledge Director that we started
                                        self.zcc.publish_ack(topic,1,1)
                                    else:
                                        self.zcc.publish_ack(topic,-1,-1)
                                    if log_msg is not None:
                                        self.zcc.publish_log(self.pub_tag, self.last_trg, self.last_trg, log_msg)
                            elif topic == 'playerlog' and new_msg[1] == self.pub_tag:
                                time.sleep(0.1)
                                self.zcc.publish_log(self.pub_tag, self.last_trg, self.last_trg, self.lastlog)
                            elif topic == 'info' and new_msg[1] == self.pub_tag:
                                time.sleep(0.05)
                                # Reload code
                                try:
                                    self.load_action_module()
                                except:
                                    log_msg = "ERROR: " + str(traceback.format_exc())
                                    self.lastlog = log_msg
                                    traceback.print_exc()
                                self.zcc.publish_info(self.prio,self.data_names,self.metadata_names)
                            else:
                                print("MAH!",new_msg)
                                self._seq_started = False
                        #
                        if time.time() - last_msg_time > 5:
                            # In the last 5 seconds we did not receive any message...
                            # check director URL
                            if not self.update_connections():
                                self._seq_started = False
                            last_msg_time = time.time()
                    except:
                        if DEBUG:
                            traceback.print_exc()
                        continue
                self._seq_started = False
                self._state = 'STANDBY'
